/* Copyright 2000, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

#include "common.h"
#include <ctype.h>

#include "adm_group.h"

/* CHECK_GROUP_NAME - Do syntax checks on a group name.  No upper case letters,
 * only a few symbols, first character must be a letter.  There is no length
 * limit for Backtalk group names.  There would be lots more restrictions on
 * unix group names, but we don't administer those.
 */

static int check_group_name(char *name)
{
static char *oksym= "_-.@";

    if (name == NULL || name[0] =='\0' || !isascii(name[0]) ||!islower(name[0]))
	return 1;

    for (name++; *name != '\0'; name++)
    {
	if (!isascii(*name) || (!islower(*name) && !isdigit(*name) &&
		!strchr(oksym,*name)))
	    return 1;
    }
    return 0;
}


/* FREE_GRP - Deallocate a single group structure.
 */

static void free_grp(struct grp *g)
{
char **m;

    if (g->name != NULL) free(g->name);
    if (g->members != NULL)
    {
	for (m= g->members; *m != NULL; m++)
	    free(*m);
    	free(g->members);
    }
    free(g);
}


/* FREE_GRPS - deallocate a whole list of groups.
 */

void free_grps(struct grp *head)
{
struct grp *g, *n;
char **m;

    for (g= head; g != NULL; g= n)
    {
	n= g->next;
	free_grp(g);
    }
}


/* APPEND_GRP - Append a group to the list whose head and tail pointers
 * are given.  The members array and name should already be in malloc'ed
 * memory.
 */

static void append_grp(struct grp **head, struct grp **tail,
    char *name, gid_t gid, char **mem)
{
    struct grp *new= (struct grp *)malloc(sizeof(struct grp));
    new->name= name;
    new->gid= gid;
    new->members= mem;
    new->next= NULL;

    if (*head == NULL)
    	*head= new;
    else
        (*tail)->next= new;

    *tail= new;
}


/* LOAD_GRPS - Load the group file into memory.  Return a pointer to the
 * head of the linked list of group structures.  Can handle arbitrarily long
 * lines, but group and user names are limited to something like 1024
 * characters.
 */

struct grp *load_grps()
{
FILE *fp;
struct grp *tail, *head= NULL;
char bf[BFSZ+1];
char *name= NULL, **mem= NULL;
int memi,memmax;
gid_t gid;
int state, ch, i, esc, j;

    if ((fp= fopen(GROUP_FILE,"r")) == NULL)
    {
    	/* Pretend we loaded the defaults */
#if defined(CFADM_GID) && defined(CFADM_GROUP)
	append_grp(&head,&tail,strdup(CFADM_GROUP),CFADM_GID,NULL);
#endif
#if defined(USER_GID) && defined(USER_GROUP)
	append_grp(&head,&tail,strdup(USER_GROUP),USER_GID,NULL);
#endif
#if defined(GRADM_GID) && defined(GRADM_GROUP)
	append_grp(&head,&tail,strdup(GRADM_GROUP),GRADM_GID,NULL);
#endif
	return head;
    }

    state= 0;
    esc= 0;
    while ((ch= fgetc(fp)) != EOF)
    {
    	switch (state)
	{
	case -1: /* skip to start of line */
	    if (ch == '\n' && !esc) state= 0;
	    break;

	case 0: /* ready to read group name */
	    if (ch == '#') {state= -1; break;}
	    state= 1;
	    i= 0;
	    /* drop through */

	case 1: /* reading group name */
	    if (ch == ':')
	    {
		bf[i]= '\0';
		if (name != NULL) free(name);
		name= strdup(bf);
		state= 10;	/* get gid next */
	    }
	    else if (ch == '\n')
	    {
		if (!esc)
		    state= 0;
		else if (i > 0)
		    i--;
	    }
	    else
	    {
		bf[i++]= ch;
		if (i >= BFSZ) state= -1;
	    }
	    break;

	case 10: /* skip to next colon (OK, the state number is out of seq) */
	    if (ch == ':') state= 2;
	    break;

	case 2: /* ready to read gid */
	    state= 3;
	    i= 0;
	    /* drop through */

	case 3: /* reading gid number */
	    if (ch == ':' || (ch == '\n' && !esc))
	    {
		bf[i]= '\0';
		/* save gid */
		gid= atoi(bf);
		state= 4;	/* get names */
	    }
	    else if (ch == '\n')	/* line ends with \ */
	    {
		if (i > 0) i--;
	    }
	    else
	    {
	    	bf[i++]= ch;
	    	if (i >= BFSZ) state= -1;
	    }
	    break;

	case 4: /* ready to read a name */
	    state= 5;
	    i= 0;

	case 5: /* reading a name */
	    if (ch == ',' || (ch == '\n' && !esc))
	    {
		if (i > 0)
		{
		    bf[i]= '\0';
		    if (mem == NULL)
		    {
			memmax= 10;
			memi= 0;
			mem= (char **)malloc(memmax*sizeof(char *));
		    }
		    else if (memi > memmax - 1)
		    {
			memmax*= 2;
			mem= (char **)realloc(mem,memmax*sizeof(char *));
		    }
		    mem[memi++]= strdup(bf);
		}

		if (ch == '\n')
		{
		    /* Line done - append it */
		    state= 0;
		    if (mem != NULL) mem[memi]= NULL;
		    append_grp(&head,&tail,name,gid,mem);
		    name= NULL;
		    mem= NULL;
		}
		else
		    state= 4;	/* read next name */
	    }
	    else if (ch == '\n')	/* line ends with \ */
	    {
		if (i > 0) i--;
	    }
	    else
	    {
	    	bf[i++]= ch;
	    	if (i >= BFSZ) state= -1;
	    }
	    break;
	}
	esc= (ch == '\\');
    }

    /* all done */
    fclose(fp);
    return head;
}


/* WRITE_GRPS - write a group structure back to the file.  If 'freemem' is true,
 * deallocate the memory for the structure.
 */

void write_grps(struct grp *head, int freemem)
{
struct grp *g, *n;
char **m;
FILE *fp;

    if ((fp= fopen(GROUP_FILE,"w")) == NULL)
    	die("cannot write to "GROUP_FILE);

    for (g= head; g != NULL; g= n)
    {
    	fprintf(fp,"%s::%d:",g->name,g->gid);
	if (g->members != NULL)
	{
	    for (m= g->members; *m != NULL; m++)
	    {
		if (m != g->members) fputc(',',fp);
		fputs(*m,fp);
		if (freemem) free(*m);
	    }
	    if (freemem) free(g->members);
	}
	fputc('\n',fp);
	n= g->next;
	if (freemem) {free(g->name); free(g);}
    }
    fclose(fp);
}


/* ADD_GRP - Add a new group (with no members) to the list.  GID is
 * automatically generated.  Returns 1 if the group name already exists.
 * Names need not be in malloc'ed memory.
 */

int add_grp(struct grp **head, char *name)
{
struct grp *g, *tail= NULL;
gid_t gid= 100;

    /* Scan through existing groups, checking that name is used, and finding
     * a free gid number */
    for (g= *head; g != NULL; tail= g, g= g->next)
    {
    	if (g->gid > gid) gid= g->gid;
    	if (!strcmp(name,g->name))
	    return 1;
    }
    gid++;
    append_grp(head,&tail,strdup(name),gid,NULL);
    return 0;
}


/* DEL_GRP - Delete the named group.  Returns 1 if it doesn't exit.
 */

int del_grp(struct grp **head, char *name)
{
struct grp *g, *p= NULL;

    for (g= *head; g != NULL; p= g, g= g->next)
    {
    	if (!strcmp(name,g->name))
	{
	    if (p == NULL)
	    	*head= NULL;
	    else
	    	p->next= g->next;
	    free_grp(g);
	    return 0;
	}
    }
    return 1;
}


/* ADD_MEMB - Add a member to an existing group on the list.
 * Returns 1 if the group name doesn't exist.
 * Names need not be in malloc'ed memory.
 */

int add_memb(struct grp *head, char *gname, char *mname)
{
struct grp *g;
char **m;
int n;

    for (g= head; g != NULL; g= g->next)
    {
    	if (!strcmp(gname,g->name))
	{
	    if (g->members == NULL)
	    {
		n= 0;
	    	g->members= (char **)malloc(2*sizeof(char *));
	    }
	    else
	    {
		/* Count current members */
		for (m= g->members, n= 0; *m != NULL; m++, n++)
		    ;
		/* Enlarge array */
		g->members= (char **)realloc(g->members,(n+2)*sizeof(char *));
	    }
	    g->members[n]= strdup(mname);
	    g->members[n+1]= NULL;
	    return 0;
	}
    }
    return 1;
}


/* DEL_MEMB - Delete a member to an existing group on the list.
 * Returns 1 if the group name doesn't exist or the member isn't in it.
 */

int del_memb(struct grp *head, char *gname, char *mname)
{
struct grp *g;
char **m;

    for (g= head; g != NULL; g= g->next)
    {
    	if (!strcmp(gname,g->name))
	{
	    if (g->members == NULL) return 1;
	    /* Count current members */
	    for (m= g->members; *m != NULL; m++)
	    {
		if (!strcmp(mname,*m))
		{
		    free(*m);
		    for (; *m != NULL; m++)
		    	m[0]= m[1];
		    m[-1]= NULL;
		    return 0;
		}
	    }
	    return 1;
	}
    }
    return 1;
}
