/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

#include "backtalk.h"

#include <time.h>
#include <signal.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#include "sysdict.h"
#include "cgi_query.h"
#include "exec.h"
#include "str.h"
#include "stack.h"
#include "dict.h"
#include "date.h"
#include "user.h"
#include "conf.h"
#include "format.h"
#include "last_log.h"
#include "byteorder.h"
#include "ps_part.h"
#include "ps_item.h"
#include "ps_sum.h"
#include "set_uid.h"


int initializing;
int ttymode= FALSE;

FILE *debugErrs= NULL;
extern char **environ;

FILE *output;			/* Where to write script output */

#ifdef REPEAT_FILE
char *load_repeat(char *fname);
void save_repeat(char *query);
void free_repeat(void);
char **old_env= NULL;
#endif /*REPEAT_FILE*/
void log(char *, char *);
#ifdef TRAP_CRASH
void set_sigtrap(void);
void unset_sigtrap(void);
#endif /*TRAP_CRASH*/

void send_favicon(char *path);

#ifdef DEBUG_MEMORY
/* Really just a place to set a breakpoint */
void merr(int errcode)
{
	printf("MEMORY ERROR: code %d\n",errcode);
}
#endif

/* Crash after a timeout */
RETSIGTYPE timeout(int arg)
{
	die("program terminated by timeout");
}


main (int argc, char **argv)
{
char *progname, *scriptdir, *id, *query= NULL;
char *file= NULL;
int i, f, rc, hflag=FALSE, saverep= TRUE;
int fd;
#ifdef TRAP_CRASH
int trapsig= TRUE;
#endif TRAP_CRASH

#ifdef DEBUG_MEMORY
	malloc_debug(2+4+8);
	malloc_error(merr);
#endif

	/* Default place to write script output */
	output= stdout;

	/* Set default umask */
	umask(022);

	/* Decide if we are big-endian or little-endian */
	check_endian();

	/* Set time limit to trap infinite loops */
	signal(SIGALRM, timeout);

#ifdef ERROR_FILE
	/* Don't create the error log file if it doesn't already exist */
	if ((fd= open(ERROR_FILE, O_WRONLY|O_APPEND)) >= 0)
		debugErrs= fdopen(fd,"a");
	else
		debugErrs= NULL;
#endif

	/* Save program name (for use in debug output) */
	if ((progname= strrchr(argv[0],'/')) == NULL)
		sets(VAR_PROGNAME,argv[0],SM_COPY);
	else
		sets(VAR_PROGNAME,progname+1,SM_COPY);


	/* Parse arguments - Most are for debugging only */
	for (i= 1; i < argc; i++)
	{
		if (argv[i][0] == '-')
			switch(argv[i][1])
			{
			case 'h':	/* Run in !ttymode even on tty */
				hflag= TRUE;
				break;
#ifdef TRAP_CRASH
			case 's':	/* Don't trap signals */
				trapsig= FALSE;
				break;
#endif
			case 'f':	/* Execute named script */
				file= argv[++i];
				ttymode= TRUE;
				break;
			case 'q':	/* Execute given query string */
				query= argv[++i];
				ttymode= TRUE;
				break;
#ifdef REPEAT_FILE
			case 'r':	/* Reexecute last query */
				ttymode= TRUE;
				saverep= FALSE;
				if (i+1 < argc)
					query= load_repeat(argv[++i]);
				else
					query= load_repeat(REPEAT_FILE);
				break;
#endif
			}
	}
	if (hflag) ttymode= FALSE;

#ifdef TRAP_CRASH
	if (trapsig)
		set_sigtrap();
#endif

	/* Find the script to execute */
	if (file == NULL) 
	{
		char *t;
		file= getenv("PATH_INFO");
		if (file != NULL)
		{
			/* Strip off leading slashes */
			while (file[0] == '/')
				file++;

			/* Don't allow paths into parent directories */
			nodotdot(file);

			/* Strip off anything after ':' for urlarg */
			if ((t= strchr(file,':')) != NULL)
			{
				*t= '\0';
				sets(VAR_URLARG,t+1,SM_COPY);
				file= strdup(file);
				*t= ':';
			}

			/* Set "flavor" to directory name */
			if ((t= strchr(file,'/')) != NULL)
			{
				char tmp= t[1];
				t[1]= '\0';
				sets(VAR_FLAVOR,file,SM_COPY);
				t[1]= tmp;
			}
		}
	}

	/* If no file given, run the welcome script instead */
	if (file == NULL || *file == '\0') file= "public/welcome";

	/* Look up and save user information */
	if ((rc= getuser()) == 2)
		file= "public/relogin";  /* script for expired sessions */

	seti(VAR_UID, user_uid());
	seti(VAR_GID, user_gid());
	sets(VAR_HOMEDIR, user_dir(), SM_COPY);
	id= user_id();
	sets(VAR_ID, id, SM_COPY);
	seti(VAR_AMADM, user_cfadm());

	/* Update the lastlog file */
	update_lastlog(idict(VAR_UID));

	/* cd into the script directory */
	if (chdir(scriptdir= strdup(sdict(VAR_SCRIPTDIR))))
	    die("script directory \"%s\" not accessible",scriptdir);

	/* Run the "config.bt" script first
	 * (initializing flag lets us redefine constants and protected vars)
	 */
	initializing= TRUE;
	run("config.bb");
	initializing= FALSE;

	/* If we don't allow anonymous, any anonymous user trying to run
	 * anything not in the 'public' directory runs public/welcome
	 */
	if (*id == '\0' && !idict(VAR_ALLOWANON) && strncmp(file,"public/",7))
	    file= "public/welcome";

	/* Remove trailing /'s from  file */
	i= strlen(file);
	while (i > 0 && file[i-1] == '/')
	    file[--i]= '\0';

	/* If a favicon.ico request, don't try to run a script */
	if (i >= 11  && !strcmp(file+i-11,"favicon.ico"))
	{
	    send_favicon(sdict(VAR_FAVICON));
	    exit(0);
	}

	/* Set mayseefname flag */
	switch (idict(VAR_ANONYMITY))
	{
	default:
	case 0: f= 1; break;
	case 1: f= (id[0] != '\0'); break;
	case 2: f= bdict(VAR_AMADM) || user_gradm(); break;
	case 3: f= bdict(VAR_AMADM); break;
	}
	seti(VAR_MAYSEEFNAME,f);

	/* cd into the new script directory, if it has changed */
	if (strcmp(scriptdir,sdict(VAR_SCRIPTDIR)) &&
	    chdir(sdict(VAR_SCRIPTDIR)))
	    die("script directory \"%s\" not accessible",sdict(VAR_SCRIPTDIR));

	/* Get the Query - if it hasn't already been set by debugging option */
	if (query == NULL) query= get_query();
#ifdef REPEAT_FILE
	if (saverep) save_repeat(query);
#endif
	if (idict(VAR_LOGLEVEL) > 0) log(file,query);
	if (query != NULL) sets(VAR_CGIQUERY,query,SM_COPY);
	load_query(query);

	/* Execute the selected script */
	run(file);

	/* Write out participation files */
	func_close_conf();

	/* Dump the stack */
	func_print();

	/* If CLEANUP is defined, deallocate everything we know about in
	 * dynamic memory.  This is done to help detect memory leaks.
	 */

#ifdef CLEANUP
	free(scriptdir);
	free_exec();			/* Free execution stack */
	free_dictstack();		/* Free dictionary stack */
	free_file_list();		/* Free file list */
	free_stack();			/* Free operand stack */
	free_cookies();			/* Free any pending cookies */
	free_timezone();		/* Free any timezone information */
	FreeTable();			/* Free text format table */
#ifdef REPEAT_FILE
	if (old_env) free_repeat();	/* free environment */
#endif /*REPEAT_FILE*/
#endif /*CLEANUP*/

	exit(0);
}


/* SEND_FAVICON - send the .ico file found in the given file.
 */

void send_favicon(char *path)
{
FILE *fp= fopen(path,"r");
int ch;

    /* Maybe this should be image/ico or image/x-ico or image/x-ms-icon or
     * any of several other things, but the normal response from most web
     * servers seems text/plain so we'll stick with that for now.
     */
    printf("Content-type: text/plain\n\n");

    if (path == NULL || path[0] == '\0' || (fp= fopen(path,"r")) == NULL)
    	return;

    while ((ch= getc(fp)) != EOF)
    	putchar(ch);
}


/* DUMP_ENV - print the environment to the given file descriptor.  If quoting
 * is true, we do HTML quoting (expanding "<" to "&lt;" and such).  If quoting
 * is false, we just turn newlines into "\n" and backslashes into "\\".
 */

void dump_env(FILE *fp, int quote)
{
char **ep, *q;

    for (ep= environ; *ep != NULL; ep++)
	if (quote)
	{
	    /* Do HTML quoting */
	    fprintf(fp,"%s\n",q= sysformat(*ep));
	    free(q);
	}
	else
	{
	    /* Expand newlines and backslashes only */
	    for (q= *ep; *q != '\0'; q++)
	    	switch (*q)
	    	{
	    	case '\n':
	    	    fputs("\\n",fp);
	    	    break;
	    	case '\\':
	    	    fputs("\\\\",fp);
	    	    break;
		default:
		    fputc(*q,fp);
		}
	    fputc('\n',fp);
	}
    fputc('\n',fp);
}


#ifdef REPEAT_FILE

/* SAVE_REPEAT - Save query and environment to a file, for possible
 * re-execution.  This is for debugging only.
 */

void save_repeat(char *query)
{
FILE *fp;

	if ((fp= fopen(REPEAT_FILE,"w")) == NULL)
	{
		fprintf(stderr,"Could not write repeat file "REPEAT_FILE"\n");
		return;
	}

	dump_env(fp, FALSE);

	if (query != NULL)
	{
		fputs(query,fp);
		fputc('\n',fp);
	}

	fclose(fp);
}

/* LOAD_REPEAT - Load a previously written repeat file.  The environment in
 * the file is stored as our environment, and the query string in the file
 * is returned.
 */

char *load_repeat(char *fname)
{
#define MALSZ 16
int esize= MALSZ, en=0;
char **newenv= malloc(esize*sizeof(char *));
static char bf[BFSZ+1];
FILE *fp;

	if ((fp= fopen(fname,"r")) == NULL)
		die("Cannot open repeat file %s to read",fname);
	
	while (fgets(bf,BFSZ,fp) != NULL)
	{
		if (en == esize)
			newenv= realloc(newenv,(esize+= MALSZ)*sizeof(char *));

		if (bf[0] == '\n')
		{
			newenv[en]= NULL;
			old_env= environ;
			environ= newenv;
			if (fgets(bf,BFSZ,fp) == NULL)
				return NULL;
			*firstin(bf,"\n")= '\0';
			return bf;
		}

		*firstin(bf,"\n")= '\0';
		unslash(bf);
		newenv[en]= (char *)malloc(strlen(bf)+1);
		strcpy(newenv[en], bf);
		en++;
	}
	die("Premature EOF in %s",fname);
}


/* FREE_REPEAT - Restore the original environment, and free up all the memory
 * used in the fake environment that we created in load_repeat().  This
 * function is normally only used when testing for memory leaks.
 */

#ifdef CLEANUP
void free_repeat()
{
int i;

	for (i= 0; environ[i] != NULL; i++)
		free(environ[i]);
	free(environ);
	environ= old_env;
}
#endif /* CLEANUP */

#endif /* REPEAT_FILE */


/* DIE - Exit with an error condition.  It prints suitable HTML headers
 * (if they haven't already been displayed), and so forth.  It also prints
 * some context information which leaks in here via various global variables,
 * plus dumps the stack.
 */

#if __STDC__ && HAVE_STDVARARGS
void die(const char *fmt,...)
#else
void die(fmt, va_alist)
    const char *fmt;
    va_dcl
#endif
{
time_t now= time(0L);
char *timestr= ctimez(&now,NULL);
#if HAVE_STDVARARGS
va_list ap;
#endif

    VA_START(ap,fmt);
    set_cfadm();

    send_http_headers();

    if (!ttymode)
    {
	fputs("<HTML><HEAD>\n<TITLE>Backtalk Crash</TITLE>\n"
	      "</HEAD><BODY BGCOLOR=FFFFFF TEXT=A00000>\n"
	      "<H1>Backtalk Crash!</H1>\n",stdout);

	printf("%24.24s<P>\n", timestr);
	fputs("<STRONG>Oops!</STRONG>  The authors of Backtalk are deeply\n"
	      "sorry to confess that our program has failed to operate\n"
	      "correctly.  This is due to an error in Backtalk itself, or\n"
	      "in one of Backtalk's script files, or in the configuration\n"
	      "of the system.  It is probably not due to anything you might\n"
	      "have done.\n", stdout);

	fputs("The page below contains an error message and a detailed dump\n"
	      "of the state Backtalk was in when the error occurred. This\n"
	      "may be useful to anyone trying to fix the problem.<P>\n",stdout);
    }

    if (debugErrs != NULL)
    {
	fprintf(debugErrs,"------- %24.24s -------\n",timestr);

	fputs("DIED:  ",debugErrs);
	VA_PRINTF(debugErrs, fmt, ap);
	fputc('\n',debugErrs);
	print_context(debugErrs);

	fputs("\n\nStack:  ",debugErrs);
	dump_stack(debugErrs,FALSE);

	putc('\n',debugErrs);
	dump_dict_stack(debugErrs);

	fputs("\nEnvironment:\n",debugErrs);
	dump_env(debugErrs, FALSE);

	if (!ttymode)
	    fputs("All this information has been logged, but it may still\n"
		  "be helpful for you to report this error and the\n"
		  "circumstances under which it occurred to the system\n"
		  "administrators.\n",stdout);
    }
    else if (!ttymode)
	fputs("This information has <STRONG>not</STRONG> been logged, so\n"
	      "it would be helpful for you to report this error to the\n"
	      "system administrators..\n",stdout);

    if (!ttymode)
	fputs("<P>Thanks for your assistance in getting the bugs out of this\n"
	      "conferencing system.  We sincerely and humbly beg your forgiveness.\n"
	      "<P><HR>\n", stdout);

    fputs(ttymode ? "\nERROR:  " : "<P>\nERROR:  ",stdout);
    VA_PRINTF(stdout, fmt, ap);
    fputs(ttymode ? "\n" : "<BR>\n",stdout);
    print_context(stdout);

    fputs(ttymode ? "\n\nStack:\n" : "<P><PRE>Stack:\n",stdout);
    dump_stack(stdout, !ttymode);

    putc('\n',stdout);
    dump_dict_stack(stdout);

    fputs(ttymode ? "\nEnvironment:\n" : "<P>\nEnvironment:\n",stdout);
    dump_env(stdout, !ttymode);

    if (!ttymode) fputs("</PRE></BODY></HTML>\n",stdout);

    /* Make sure we zap any lock files we may have */
    closepart();
    closeitem();
    closesum();

    VA_END(ap);
    exit(0);
}


/* DIE_NOSCRIPT - Exit error condition.  This is used when attempting to run
 * non-existant scripts.  It creates a abbreviated and less apologetic crash
 * screen, because this is usually not our fault.
 */

void die_noscript(char *file)
{
time_t now= time(0L);
char *timestr= ctimez(&now,NULL);
char *e;

    set_cfadm();

    send_http_headers();

    if (!ttymode)
    {
	fputs("<HTML><HEAD>\n<TITLE>Backtalk Script Not Found</TITLE>\n"
	      "</HEAD><BODY BGCOLOR=FFFFFF TEXT=A00000>\n"
	      "<H1>Backtalk Script Not Found</H1>\n",stdout);

	printf("%24.24s<P>\n", timestr);
    }

    if (debugErrs != NULL)
    {
	fprintf(debugErrs,"------- %24.24s -------\n",timestr);

	fprintf(debugErrs,"DIED:  Script %s does not exist\n",file);
	print_context(debugErrs);

	fprintf(debugErrs,"\ncgiquery=%s\n",sdict(VAR_CGIQUERY));
	fprintf(debugErrs,"REQUEST_URI=%s\n",(e= getenv("REQUEST_URI"))?e:"");
	fprintf(debugErrs,"PATH_INFO=%s\n",(e= getenv("PATH_INFO"))?e:"");
	fprintf(debugErrs,"QUERY_STRING=%s\n",(e= getenv("QUERY_STRING"))?e:"");
	fprintf(debugErrs,"REMOTE_ADDR=%s\n",(e= getenv("REMOTE_ADDR"))?e:"");
	fprintf(debugErrs,"REMOTE_PORT=%s\n",(e= getenv("REMOTE_PORT"))?e:"");
	fprintf(debugErrs,"REMOTE_USER=%s\n",(e= getenv("REMOTE_USER"))?e:"");
    }

    printf("You have attempted to run a Backtalk script named '%s'.\n"
	    "There is no Backtalk script by this name.\n"
	    "If you typed the URL in yourself, then you entered it\n"
	    "incorrectly.  If you got this message after clicking a link,\n"
	    "then that link was erroneous.  If the link was on a Backtalk\n"
	    "page, then probably Backtalk is broken.  Please inform the\n"
	    "system administrators if this is the case.\n",file);

    if (!ttymode) fputs("<PRE>\n",stdout);

    printf("cgiquery=%s\n",sdict(VAR_CGIQUERY));
    printf("REQUEST_URI=%s\n",(e= getenv("REQUEST_URI"))?e:"");
    printf("PATH_INFO=%s\n",(e= getenv("PATH_INFO"))?e:"");
    printf("QUERY_STRING=%s\n",(e= getenv("QUERY_STRING"))?e:"");
    printf("REMOTE_ADDR=%s\n",(e= getenv("REMOTE_ADDR"))?e:"");
    printf("REMOTE_PORT=%s\n",(e= getenv("REMOTE_PORT"))?e:"");
    printf("REMOTE_USER=%s\n",(e= getenv("REMOTE_USER"))?e:"");

    if (!ttymode) fputs("</PRE></BODY></HTML>\n",stdout);

    /* Make sure we zap any lock files we may have */
    closepart();
    closeitem();
    closesum();

    VA_END(ap);
    exit(0);
}


/* LOG - write a log entry for the current transaction.  Don't call if loglevel
 * is zero.  The original plan was to output more information with higher
 * loglevels, but there doesn't really seem to be all that much interesting
 * information to output.
 */

void log(char *file, char *query)
{
time_t clock= time(0L);
char *datum;
FILE *fp;

	if ((fp= fopen(LOG_FILE,"a")) == NULL)
		return;
	
	fprintf(fp,"%15.15s %s", ctimez(&clock,NULL)+4, file);

        if ((datum= sdict(VAR_ID)) != NULL && datum[0] != '\0')
                fprintf(fp, " %s", datum);
	else
		fprintf(fp, " anonymous");
        if ((datum= getenv("REMOTE_HOST")) != NULL)
                fprintf(fp, " %s", datum);
        if ((datum= getenv("REMOTE_ADDR")) != NULL)
                fprintf(fp, "(%s)", datum);

	if (query != NULL)
		fprintf(fp,"\n %s\n", query);
	else
		fputc('\n',fp);
}

#ifdef TRAP_CRASH

/* SIGTRAP - This routine is used to catch weird signals, like segmentation
 * violations, that would normally cause crashes and core dumps.  It tries
 * to generate a backtalk crash screen instead of a core dump, on the
 * assumption that in a CGI environment the former is more useful.  In many
 * cases this won't work, since whatever caused the crash may have screwed
 * things up too badly to be able to even die gracefully, but it's worth
 * a try.
 */

RETSIGTYPE sigtrap(int arg)
{
	/* Could skip unset on systems with oneshot signal signal handlers */
	unset_sigtrap();

	die("terminated by signal %d",arg);
}


/* SET_SIGTRAP - set traps on most signals that could kill the process.
 */

void set_sigtrap()
{
#ifdef SIGABRT
	signal(SIGABRT,sigtrap);
#endif
#ifdef SIGBUS
	signal(SIGBUS,sigtrap);
#endif
#ifdef SIGFPE
	signal(SIGFPE,sigtrap);
#endif
#ifdef SIGHUP
	signal(SIGHUP,sigtrap);
#endif
#ifdef SIGILL
	signal(SIGILL,sigtrap);
#endif
#ifdef SIGINT
	signal(SIGINT,sigtrap);
#endif
#ifdef SIGPIPE
	signal(SIGPIPE,sigtrap);
#endif
#ifdef SIGPWR
	signal(SIGPWR,sigtrap);
#endif
#ifdef SIGSEGV
	signal(SIGSEGV,sigtrap);
#endif
#ifdef SIGSYS
	signal(SIGSYS,sigtrap);
#endif
/* Don't trap SIGTERM - http servers send them often and we should honor them
#ifdef SIGTERM
	signal(SIGTERM,sigtrap);
#endif
*/
#ifdef SIGUSR1
	signal(SIGUSR1,sigtrap);
#endif
#ifdef SIGUSR2
	signal(SIGUSR2,sigtrap);
#endif
#ifdef SIGQUIT
	signal(SIGQUIT,sigtrap);
#endif
#ifdef SIGXCPU
	signal(SIGXCPU,sigtrap);
#endif
#ifdef SIGXFSZ
	signal(SIGXFSZ,sigtrap);
#endif
}


/* UNSET_SIGTRAP - set traps on most signals that could kill the process.
 */

void unset_sigtrap()
{
#ifdef SIGABRT
	signal(SIGABRT,SIG_DFL);
#endif
#ifdef SIGBUS
	signal(SIGBUS,SIG_DFL);
#endif
#ifdef SIGFPE
	signal(SIGFPE,SIG_DFL);
#endif
#ifdef SIGHUP
	signal(SIGHUP,SIG_DFL);
#endif
#ifdef SIGILL
	signal(SIGILL,SIG_DFL);
#endif
#ifdef SIGINT
	signal(SIGINT,SIG_DFL);
#endif
#ifdef SIGPIPE
	signal(SIGPIPE,SIG_DFL);
#endif
#ifdef SIGPWR
	signal(SIGPWR,SIG_DFL);
#endif
#ifdef SIGSEGV
	signal(SIGSEGV,SIG_DFL);
#endif
#ifdef SIGSYS
	signal(SIGSYS,SIG_DFL);
#endif
#ifdef SIGTERM
	signal(SIGTERM,SIG_DFL);
#endif
#ifdef SIGUSR1
	signal(SIGUSR1,SIG_DFL);
#endif
#ifdef SIGUSR2
	signal(SIGUSR2,SIG_DFL);
#endif
#ifdef SIGQUIT
	signal(SIGQUIT,SIG_DFL);
#endif
#ifdef SIGXCPU
	signal(SIGXCPU,SIG_DFL);
#endif
#ifdef SIGXFSZ
	signal(SIGXFSZ,SIG_DFL);
#endif
}

#endif /*TRAP_CRASH*/
